// functions.cpp V.1.0
// Date: 29/II/2016
// 
// This file is part of the "Erik+2" program
/*
This code has been written by M. Casanellas and J. Fernandez-Sanchez. 
Please quote the paper "Invariant versus classical quartet inference when evolution is heterogeneous across sites and lineages", Sys. Bio. doi:10.1093/sysbio/syv086 .
*/

#include "declaration.h"
#include <stdlib.h>
#include <stdio.h>
#include <algorithm>
#include <armadillo>
#include <string>
#include <fstream>
#include <iostream>


using namespace std;	//using namespace std;
using namespace arma; // for 'armadillo';

string nuc[4]={"A","C","G","T"};
int N=4;
int numb_sp;
vector <string> observations=createpatterns(4);
vector <int> power4; 


vector<int> get_species()
{
	int aux=1; 
	int numb_sp=4; 
	for (int i=0;i<=numb_sp;i++) {	
		power4.push_back(aux); 
		aux=4*aux;
	}
	return(power4); 
}

Alignment readFASTA (string fname) { 	
	Alignment align;
	string line;
	// vector<string> species;
  fstream file;
  string speciesname;
  string dna="";
  string dnaaux;
  int numchar;
  int i;
  vector<int> al_length;
  //opening file
  file.open(fname.c_str(), fstream::in); //open(pointer to filename, type i.e. in/out)
 	if (file == NULL) {
 		cout <<"cannot open file " << fname.c_str() << "\n";
 		exit(1);
	}
  // get the first line from the FASTA file
  getline(file, line, '\n');//reads the first line
  if (line[0]!='>'){
	 	cout <<"error: not a FASTA file \n";
  	exit(0);
  }
  numchar=line.length()-1;
  speciesname=line.substr(1,numchar);
  align.taxa.push_back(speciesname);
  if (align.seqs[speciesname].length() > 0) {
		cerr << "Warning - found 2+ sequences with same name " << speciesname << endl;		
	}
	while(! file.eof()){
		getline(file, line, '\n');
	  
		if (line[0]!='>'){
   		dna += line;
   	}
		else{ 
   		align.seqs[speciesname]=dna;
   		al_length.push_back(dna.length());
   		numchar=line.length()-1;
   		speciesname=line.substr(1,numchar);
   		align.taxa.push_back(speciesname);
   		dna.clear();
   		if (align.seqs[speciesname].length() > 0) {
				cerr << "Warning - found 2+ sequences with same name " << speciesname << endl;		
			}
		}
	}
	align.seqs[speciesname]=dna;
	al_length.push_back(dna.length());
	file.close();
	align.num_taxa=align.taxa.size();
	if(al_length.size() != align.num_taxa)	cout << "error in getting the alignment!!!";
	for (i=0; i< al_length.size()-1; i++){
		if (al_length[i] != al_length[i+1]){
	   	cout << "sequence " <<align.taxa[i+1] <<" doesn't have same length \n";
	   	exit(0);
 		}
	}
	align.seq_len=al_length.back();
	return align;
}



vector<string> makepatterns(vector<string> &patterns){
	string patt;
  int s=patterns.size();
  vector<string>::iterator it;
	for (int j=0;j<s; j++){
   	patt=patterns[0];

		patterns.push_back(patt.append("A"));
  	unsigned found = patt.find_last_of("A");
		patt=patt.substr(0,found);
		// patt.pop_back();
		patterns.push_back(patt.append("C"));
		found = patt.find_last_of("C");
		patt=patt.substr(0,found);
		// patt.pop_back();
		patterns.push_back(patt.append("G"));
		found = patt.find_last_of("G");
		patt=patt.substr(0,found);
		// patt.pop_back();
		patterns.push_back(patt.append("T"));
	  patterns.erase(patterns.begin());
	}
	return(patterns);
}


 // from an alignment, it obtains the relative frequencies of the columns
 map<string, int> Getcolumns (Alignment &align){
	int i,numb_taxa,al_length=align.seq_len; 
	long long int numb_obs;
	float aux=1./al_length; 
	long int numb_col;
	string col;

	numb_taxa=align.num_taxa;

	map<string,string> myseqs=align.seqs;
	map<string,string>::const_iterator pos;
	map<long int,int>::const_iterator pattern;
	map<string, int> columnscount;
	string mystring;

	numb_obs=power4[numb_taxa];

	for (i=0; i < al_length; i++) {
		col="";

	for (int j=0;j<align.num_taxa;j++) {	
		mystring=myseqs[align.taxa[j]];
		col+=mystring[i];
	}

	/*	for (pos = myseqs.begin(); pos != myseqs.end(); ++pos) {
     			col.push_back(pos->second[i]);
	}*/
	columnscount[col]++;
	}
	return (columnscount);
}

// it generates all possible observations at the leaves of a tree with n leaves
vector<string> createpatterns(int n) {
 	vector<string> patterns;
 	patterns.push_back("A");
 	patterns.push_back("C");
 	patterns.push_back("G");
 	patterns.push_back("T");

 	for (int i=0;i < n-1; i++)	patterns=makepatterns(patterns);
 	return(patterns);
}


float Erik2(mat flat, int n_partitions, vector<int> split, vector<int> count_split)
{
	int i,k,j,a,b;
	mat flat_colcorrected,flat_rowcorrected;
	mat colsum,rowsum;
	int obs_row=flat.n_rows; 
	int obs_col=flat.n_cols; 
	vector<int> cols2remove; 
	vector<int> rows2remove; 

	a=split.size();
	b=count_split.size();
	mat unit_split = ones<mat>(1,obs_row);
	mat unit_othersp = ones<mat>(obs_col,1);
	vec singval_col, singval_row;
	colsum=unit_split*flat; 
	rowsum=flat*unit_othersp; 

	for (i=0;i<colsum.size();i++) {
		if (colsum[i]<=2) cols2remove.push_back(i); 
		else {colsum[i]=1./colsum[i];}
	}
	for (i=0;i<rowsum.size();i++) {
		if (rowsum[i]<=2) rows2remove.push_back(i); 
		else {rowsum[i]=1./rowsum[i];}
	}
	int n_cols2remove=cols2remove.size();
	int n_rows2remove=rows2remove.size();
	flat_colcorrected=flat*diagmat(colsum);
	flat_rowcorrected=diagmat(rowsum)*flat;
	for(j=n_cols2remove-1;j>=0;j--) {
		flat_colcorrected.shed_col(cols2remove[j]);
	}
	// cout << "col corrected " << endl; 
	// cout << flat_colcorrected << endl; 
	for(j=n_rows2remove-1;j>=0;j--) {
		flat_rowcorrected.shed_row(rows2remove[j]);
	}
	// flat_rowcorrected.shed_col(1);
	// cout << "row corrected " << endl; 
	// cout << flat_rowcorrected << endl; 

	singval_col=svd(flat_colcorrected);	
	// cout << "singular values columns: " << singval_col << endl; 
	singval_row=svd(flat_rowcorrected); 
	// cout << "singular values rows: " << singval_row << endl; 

	float tail_row=0;
	float tail_col=0;
	for(k=4*n_partitions;k<singval_col.n_elem;k++) {
		tail_col+=singval_col[k]*singval_col[k];
	}
	for(k=4*n_partitions;k<singval_row.n_elem;k++) {
		tail_row+=singval_row[k]*singval_row[k];
	}
	return((sqrt(tail_col)+sqrt(tail_row))/2);
}



mat Flattening(map<string,int> tensor, vector<int> split, vector <int> othersp)
{
 	int i,j,k,a,b; 
	long int coord_row,coord_col;
	map<long int,map<long int,int>> matrix;
	vector<long int> coord_it; 
	map<long int,int> mat_rows, mat_cols;
	int i_row=0;
	int i_col=0;

	for (map<string,int>::iterator it=tensor.begin(); it!=tensor.end(); ++it) {
		coord_it=Decomp(it->first,split,othersp);
		coord_row=coord_it[0];
		coord_col=coord_it[1];
		matrix[coord_row][coord_col]=it->second;

		if (mat_rows.find(coord_row)==mat_rows.end()) {
			i_row++; 
			mat_rows[coord_row]=i_row-1;
		}			
		if (mat_cols.find(coord_col)==mat_cols.end()) {
			i_col++; 
			mat_cols[coord_col]=i_col-1;
		}
	}
	map<long int,int>::iterator index_i;
	map<long int,int>::iterator index_j;

	long int n_rows=mat_rows.size(); 
	long int n_cols=mat_cols.size();
	mat flatmatrix=zeros<mat>(n_rows,n_cols);
	for (index_i=mat_rows.begin();index_i!=mat_rows.end();++index_i) {
		for (index_j=mat_cols.begin();index_j!=mat_cols.end();++index_j) {		
			flatmatrix(index_i->second,index_j->second)=matrix[index_i->first][index_j->first];
		}
	}
	return flatmatrix;
}


vector<long int> Decomp(string it, vector <int> split, vector <int> othersp) 
{
	int i; 
	int aa=split.size();
	int bb=othersp.size();;
	int numb=aa+bb;

	long int coord_row=0;
	long int coord_col=0;

	vector <long int> coords; 
	for(i=0;i<aa;i++) {
		coord_row+=(it[split[i]]-'0')*power4[numb-1-split[i]];
	}	
	for(i=0;i<bb;i++) {
		coord_col+=(it[othersp[i]]-'0')*power4[numb-1-othersp[i]];
	}	
	coords={coord_row,coord_col};
	return(coords);	
}


string convertInt(int number)
{
    if (number == 0) return "0";
    string temp="";
    string returnvalue="";
    while (number>0) 
	 {
        temp+=number%10+48;
        number/=10;
    }
    for (int i=0;i<temp.length();i++) returnvalue+=temp[temp.length()-i-1];
    return returnvalue;
}


